use hidden::HList_;

mod hidden {
    pub trait HList_ {}
}

pub trait Reversed<Cont>
{
    type Output: HList_;
    fn reversed_prepend(self, cont: Cont) -> Self::Output;
}

impl<Cont> Reversed<Cont> for Nil
    where Cont: HList_
{
    type Output = Cont;

    fn reversed_prepend(self, cont: Cont) -> Self::Output {
        cont
    }
}

impl<Cont, Item, Rest> Reversed<Cont> for Cons<Item, Rest>
    where 
        Cont: HList_,
        Rest: HList_ + Reversed<Cons<Item, Cont>>,
        Self: HList_,
        Cons<Item, Cont>: HList_
{
    type Output = <Rest as Reversed<Cons<Item, Cont>>>::Output;

    fn reversed_prepend(self, cont: Cont) -> Self::Output {
        self.1.reversed_prepend(Cons(self.0, cont))
    }

}

pub trait HList
    where 
        Self: Sized + HList_ + Reversed<Nil> {
    fn prepend<T>(self, value: T) -> Cons<T, Self> {
        Cons(value, self)
    }

    type Reversed;
    fn reversed(self) -> Self::Reversed;
}
impl HList for Nil {

    type Reversed = Self::Output;

    fn reversed(self) -> Nil{
        Nil
    }
}
impl<Item, Rest> HList for Cons<Item, Rest> 
    where 
        Rest: HList_,
        Self: Reversed<Nil>
{
    type Reversed = Self::Output;
    fn reversed(self) -> Self::Output {
        self.reversed_prepend(Nil)
    }
}

#[derive(Debug, Default, Clone, Copy, PartialEq, PartialOrd, Ord, Eq, Hash)]
pub struct Cons<Item, Rest: HList_>(pub Item, pub Rest);
impl<Item, Rest> HList_ for Cons<Item , Rest> 
    where Rest: HList_ {}
impl<Item, Rest: HList_> Cons<Item, Rest> {
    pub fn uncons(self) -> (Item, Rest) {
        (self.0, self.1)
    }

    pub fn head(&self) -> &Item {
        &self.0
    }
}

#[derive(Debug, Default, Clone, Copy)]
pub struct Nil;
impl HList_ for Nil {}
